/*
 * Copyright © 2024-2025 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of Nvidia Corporation and its affiliates
 * (the "Company") and all right, title, and interest in and to the software
 * product, including all associated intellectual property rights, are and
 * shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#define _GNU_SOURCE

#define ENABLE_EV 0

#if ENABLE_EV
#include <ev.h>
#endif

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if EXTERNAL_TEST
#include "clx_api.h"                       // NOLINT
#include "clx_api_generate_random_word.h"  // NOLINT
#include "clx_api_metrics.h"               // NOLINT
#else
#include "api/clx_api.h"
#include "api/clx_api_generate_random_word.h"
#include "api/clx_api_metrics.h"
#include "api/clx_api_metrics_compare.h"
#if CLX_WITH_GRPC
#include "data/prom_reader/clx_read_metrics_from_prometheus.h"
#endif
#endif

#define RANDOM_SEED                0
#define ENABLE_FILE_WRITE          1
#define ENABLE_SERIALIZE_TO_BINARY 0
#define ENABLE_TEST_HISTOGRAM      0
#define MERGE_TIMESTAMPS           1

#define DATA_PAGE_SIZE 1024 * 64  // 64K
// #define DATA_PAGE_SIZE  3 * 512;  // 1.5K

#define COUNTERS_NUMBER 499
// #define COUNTERS_NUMBER 1
#define REPEATS_NUMBER 10000

#define LABEL_SETS_NUMBER   4
#define LABEL_VALUES_NUMBER 3  // max is 4 in this example

#if ENABLE_SERIALIZE_TO_BINARY
#if EXTERNAL_TEST
#include "clx_api_metrics_to_binary_block.h"  // NOLINT
#else
#include "api/clx_api_metrics_to_binary_block.h"
#endif
#endif

static uint64_t the_counter_value = 0;

#if ENABLE_SERIALIZE_TO_BINARY
static void hexdump(FILE* out, const void* data, size_t size) {
    const unsigned char* ptr  = (const unsigned char*)data;
    const size_t bytesPerLine = 16;

    for (size_t i = 0; i < size; i += bytesPerLine) {
        // Print offset
        fprintf(out, "%08zu: ", i);
        // Print hex values
        for (size_t j = 0; j < bytesPerLine; ++j) {
            if (i + j < size) {
                fprintf(out, "%02x ", ptr[i + j]);
            } else {
                fprintf(out, "   ");
            }
        }
        fprintf(out, " ");
        // Print ASCII representation
        for (size_t j = 0; j < bytesPerLine; ++j) {
            if (i + j < size) {
                unsigned char c = ptr[i + j];
                fprintf(out, "%c", (c >= 32 && c <= 126) ? c : '.');
            }
        }
        fprintf(out, "\n");
    }
}
#endif

#if ENABLE_TEST_HISTOGRAM
static unsigned int random_uint(unsigned int from, unsigned int to) {
    if (from > to) {
        unsigned int temp = from;
        from              = to;
        to                = temp;
    }

    unsigned int range      = to - from + 1;
    unsigned int random_num = (unsigned int)rand() % range + from;

    return random_num;
}

static void test_histogram(clx_metrics_api_context_t* ctx) {
    const char* label_names[]  = {"lbl_a", "lbl_b", "lbl_c"};
    const char* label_values[] = {"est", "ost"};
    double buckets[2]          = {0.5, 1.0};

    uint64_t timestamp = clx_api_get_timestamp();

    size_t label_set_id = clx_api_metrics_add_label_names(ctx, label_names, 3);
    if (label_set_id == CLX_METRICS_API_LABEL_SET_ID_IS_INVALID) {
        return;
    }

    clx_metrics_api_histogram_id_t id =
        clx_api_metrics_add_histogram(ctx, timestamp, "wind", buckets, 2, label_set_id, label_values, 0);

    clx_api_metrics_histogram_observe(ctx, id, 0.3);
    clx_api_metrics_histogram_observe(ctx, id, 0.7);
    // clx_api_metrics_histogram_observe(ctx, id, 1.5);

    usleep(random_uint(1100, 1200));
    clx_api_metrics_histogram_flush(ctx, id, CLX_METRICS_API_HISTOGRAM_TIMESTAMP_UPDATE, false);

    clx_api_metrics_flush(ctx);

    clx_api_metrics_histogram_observe(ctx, id, 1.5);
    usleep(random_uint(1100, 1200));
    clx_api_metrics_histogram_flush(ctx, id, CLX_METRICS_API_HISTOGRAM_TIMESTAMP_UPDATE, false);
}
#endif

#if ENABLE_SERIALIZE_TO_BINARY
static void serialize_entire_context(clx_metrics_api_context_t* ctx, bool parse_binary_block) {
    size_t block_size           = 0;
    unsigned char* binary_block = clx_api_metrics_serialize_to_binary_block(ctx, &block_size);

    printf("Hexdump of Binary Block:\n");
    hexdump(stdout, binary_block, block_size);

    if (parse_binary_block) {
        clx_api_metrics_parse_full_block(ctx, binary_block, block_size);
    }
    clx_api_metrics_free_binary_block(binary_block);
}
#endif

static int test_clx_metrics_api(clx_metrics_api_context_t* ctx, int repeats, bool merge_timestamps) {
#if RANDOM_SEED
    unsigned int seed = (unsigned int)time(NULL);
#else
    unsigned int seed = 1000;
#endif

    // Generate LABEL_SETS_NUMBER sets of label_names
    const char* label_sets[LABEL_SETS_NUMBER][4];
    for (size_t i = 0; i < LABEL_SETS_NUMBER; ++i) {
        for (size_t j = 0; j < 4; ++j) {
            label_sets[i][j] = clx_api_get_random_noun(seed);
        }
    }

    uint64_t total_delay = 0;
    uint64_t timestamp   = clx_api_get_timestamp();
    // int timestamp_update_countdown                     = 0;
    static clx_metrics_api_histogram_id_t histogram_id = -1;

    if (histogram_id == CLX_METRICS_API_HISTOGRAM_ID_IS_INVALID) {
        size_t label_set_id = clx_api_metrics_add_label_names(ctx, label_sets[0], 4);
        if (label_set_id == CLX_METRICS_API_LABEL_SET_ID_IS_INVALID) {
            return -1;
        }
#if 0
        const char* label_values[] = {"hist_label_A", "hist_label_B", "hist_labelC"};
        double buckets[3]          = {50100.11 + repeats, 100100.22 + repeats, 150100.33 + repeats};
        histogram_id = clx_api_metrics_add_histogram(ctx, timestamp, "voltage", buckets, 3, label_set_id, label_values, 0);

        clx_api_metrics_histogram_observe(ctx, histogram_id, 20000);
        for (size_t i = 0; i < 3; i++) {
            clx_api_metrics_histogram_observe(ctx, histogram_id, 50000 + 200 * i);
        }
        clx_api_metrics_histogram_observe(ctx, histogram_id, 11000);
        clx_api_metrics_histogram_observe(ctx, histogram_id, 61000);
        clx_api_metrics_histogram_observe(ctx, histogram_id, 71000);
#endif
    }

    clx_api_add_constant_label(ctx, "source", "source_value");
    clx_api_add_constant_label(ctx, "tag", "tag_value");

    static char counter_names[COUNTERS_NUMBER][20];
    for (int i = 0; i < COUNTERS_NUMBER; i++) {
        snprintf(counter_names[i], sizeof(counter_names[i]), "cntr_%03d", i);
    }

    timestamp = clx_api_get_timestamp();

    for (int repeat = 0; repeat < repeats; repeat++) {
        size_t idx = 0;
        for (size_t i = 0; i < LABEL_SETS_NUMBER; ++i) {
            size_t label_set_id = clx_api_metrics_add_label_names(ctx, label_sets[i], 4);

            idx++;

            char lbl_values[LABEL_VALUES_NUMBER][20];
            const char* label_values[LABEL_VALUES_NUMBER];
            for (size_t j = 0; j < LABEL_VALUES_NUMBER; ++j) {
                snprintf(lbl_values[j], sizeof(lbl_values[j]), "%lu", 1000 * idx + j);
                label_values[j] = lbl_values[j];
            }

            for (size_t c = 0; c < COUNTERS_NUMBER; ++c) {
                the_counter_value++;

                uint64_t counter_value = the_counter_value;

                uint64_t ts = clx_api_get_timestamp();
                if (merge_timestamps) {
                    if (ts - timestamp > 1000) {
                        timestamp = ts;
                    }
                } else {
                    timestamp = ts;
                }

                // size_t counter_label_set_id = random_uint(0, 3);
                size_t counter_label_set_id = label_set_id;
                clx_api_metrics_add_counter(ctx, timestamp, counter_names[c], counter_value, counter_label_set_id, label_values,
                                            LABEL_VALUES_NUMBER);
#if 1
                if (histogram_id == CLX_METRICS_API_HISTOGRAM_ID_IS_INVALID) {
                    // uint64_t timestamp = clx_api_get_timestamp();
                    double buckets[3] = {50100.11, 100100.22, 150100.33};
                    histogram_id      = clx_api_metrics_add_histogram(ctx, timestamp, "histo_name1", buckets, 3, label_set_id,
                                                                      label_values, LABEL_VALUES_NUMBER);
                }
                clx_api_metrics_histogram_observe(ctx, histogram_id, counter_value);
#endif
            }
        }
        uint64_t elapsed = clx_api_get_timestamp() - timestamp;
        if (merge_timestamps && elapsed < 1001) {
            uint64_t delay = 1001 - elapsed;
            usleep(delay);  // add delay to avoid duplicate metric values with the same timestamp
            total_delay += delay;
        }
        if (repeat % 1000 == 0) {
            printf("total delay: %" PRIu64 " ms\n", total_delay / 1000);
        }
    }

    if (histogram_id != CLX_METRICS_API_HISTOGRAM_ID_IS_INVALID) {
        clx_api_metrics_histogram_flush(ctx, histogram_id, CLX_METRICS_API_HISTOGRAM_TIMESTAMP_NO_CHANGE, true);
    }

    return 0;
}

static void init_clx_api_params(clx_api_params_t* params) {
    // params->buffer_size = 0;  // use default
    params->buffer_size = DATA_PAGE_SIZE;

#if ENABLE_FILE_WRITE
    params->max_file_size      = 1 * 1024 * 1024;     // 1 MB
    params->max_file_age       = 60 * 60 * 1000000L;  // 1 hour
    params->schema_path        = "tmp/example/schema";
    params->data_root          = "tmp/example";
    params->source_id          = "S1";
    params->source_tag         = "example_event";
    params->file_write_enabled = true;
    params->data_path_template = CLX_API_DATA_PATH_TEMPLATE;
#endif

    char* remote_write_receiver = getenv("CLX_REMOTE_WRITE_RECEIVER");
    if (remote_write_receiver) {
        params->enabled_exporters = "remote_write_exporter";
    }
    char* open_telemetry_receiver = getenv("CLX_OPEN_TELEMETRY_RECEIVER");
    if (open_telemetry_receiver) {
        params->enabled_exporters = "open_telemetry_exporter";
    }
    char* lua_script_dir = getenv("CLX_LUA_SCRIPT_DIR");
    if (lua_script_dir) {
        params->enabled_exporters = "rta_exporter";
    }
}

static bool test_event_provider_initialize(clx_api_context_t* ctx, clx_api_provider_t* provider) {
    (void)ctx;
    (void)provider;
    return true;
}

static clx_api_provider_t* get_provider(void) {
    clx_api_provider_t* provider = calloc(1, sizeof(clx_api_provider_t));
    clx_api_version_t ver        = {{1, 0, 0}};
    provider->name               = strdup("My_provider");
    provider->version            = ver;
    provider->initialize         = &test_event_provider_initialize;

    // can be used for additional provider-specific info
    return provider;
}

static void destroy_provider(clx_api_provider_t* provider) {
    free(provider->name);
    free(provider);
}

#if ENABLE_EV

// Callback function for the timer
static void timeout_cb(EV_P_ ev_timer* w, int revents) { printf("Waiting for ^C\n"); }

static void stop_cb(EV_P_ ev_timer* w, int revents) {
    // printf("5 minutes have passed. Stopping the event loop.\n");
    ev_break(EV_A_ EVBREAK_ALL);
}

static void run_timer(double secs) {
    struct ev_loop* event_loop = ev_loop_new(EVFLAG_AUTO);
    // Initialize a timer watcher
    ev_timer timeout_watcher;
    ev_timer_init(&timeout_watcher, timeout_cb, 0., 10.);  // Trigger every 10 second
    ev_timer_start(event_loop, &timeout_watcher);

    // Initialize a stop timer watcher
    ev_timer stop_watcher;
    ev_timer_init(&stop_watcher, stop_cb, secs, 0.);  // Stop after 300 seconds (5 minutes)
    ev_timer_start(event_loop, &stop_watcher);

    ev_run(event_loop, 0);
    ev_loop_destroy(event_loop);
}
#endif
int main(int argc, char* argv[]) {
    (void)argc;
    (void)argv;

    //  Define params
    clx_api_params_t params = {0};
    init_clx_api_params(&params);

    clx_api_provider_t* provider = get_provider();
    if (provider == NULL) {
        printf("Failed to initialize provider\n");
        return -1;
    }

    clx_api_context_t* clx_api_ctx = clx_api_create_context(&params, provider);
    clx_metrics_api_context_t* ctx = clx_api_metrics_create_context(clx_api_ctx);

#if RANDOM_SEED
    unsigned int seed = (unsigned int)time(NULL);
#else
    unsigned int seed = 1735902662;
#endif

    srand(seed);

#if ENABLE_TEST_HISTOGRAM
    test_histogram(ctx);
#endif

#if ENABLE_EV
    run_timer(2);
#else
    sleep(2);
#endif

    int result = 0;

    int repeats = REPEATS_NUMBER;
    result      = test_clx_metrics_api(ctx, repeats, MERGE_TIMESTAMPS);

#if ENABLE_SERIALIZE_TO_BINARY
    // Serialize the entire context
    serialize_entire_context(ctx, params.enabled_exporters == NULL);
#endif

    // flush
    clx_api_metrics_flush(ctx);

    // waiting for remote write completion
#if ENABLE_EV
    run_timer(5);
#else
    sleep(5);
#endif

#if EXTERNAL_TEST
    clx_api_metrics_destroy_context(ctx);
    clx_api_destroy_context(clx_api_ctx);
    destroy_provider(provider);
#else

    clx_metrics_api_compare_context_t* cmp_ctx = clx_api_metrics_compare_create_context(ctx);

    clx_api_metrics_destroy_context(ctx);
    clx_api_destroy_context(clx_api_ctx);
    destroy_provider(provider);

#if ENABLE_FILE_WRITE
    sleep(3);
#if CLX_WITH_GRPC
    const char* endpoint = getenv("CLX_API_METRICS_ROUNDTRIP_CHECK_ENDPOINT");
    if (endpoint) {
        clx_api_metrics_compare(cmp_ctx, endpoint, clx_read_metrics_from_prometheus);
    }
#endif
#endif
    clx_api_metrics_compare_destroy_context(cmp_ctx);
#endif

    return result;
}
